"""
Tests for shillelagh.functions.
"""

import json
import sys

import apsw
import pytest
from pytest_mock import MockerFixture

from shillelagh.adapters.registry import AdapterLoader
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError
from shillelagh.functions import date_trunc, get_metadata, upgrade

from .fakes import FakeAdapter

if sys.version_info < (3, 10):
    from importlib_metadata import distribution
else:
    from importlib.metadata import distribution


def test_sleep_from_sql(mocker: MockerFixture, registry: AdapterLoader) -> None:
    """
    Test ``sleep``.
    """
    sleep = mocker.patch("time.sleep")
    registry.add("dummy", FakeAdapter)
    connection = connect(":memory:", ["dummy"], isolation_level="IMMEDIATE")
    cursor = connection.cursor()
    cursor.execute("SELECT sleep(5)")

    sleep.assert_called_with(5)


def test_get_metadata() -> None:
    """
    Test ``get_metadata``.
    """
    assert (
        get_metadata(
            {"dummy": {"key": "value"}, "other": {"one": "two"}},
            [FakeAdapter],
            "dummy://",
        )
        == '{"extra": {}, "adapter": "FakeAdapter"}'
    )

    with pytest.raises(ProgrammingError) as excinfo:
        get_metadata({}, [], "dummy://")
    assert str(excinfo.value) == "Unsupported table: dummy://"

    with pytest.raises(ProgrammingError) as excinfo:
        get_metadata({}, [FakeAdapter], "invalid://")
    assert str(excinfo.value) == "Unsupported table: invalid://"


def test_get_metadata_from_sql(mocker: MockerFixture, registry: AdapterLoader) -> None:
    """
    Test calling ``get_metadata`` from SQL.
    """
    mocker.patch(
        "shillelagh.functions.get_metadata",
        return_value=json.dumps({"hello": "world"}),
    )
    registry.add("dummy", FakeAdapter)
    connection = connect(":memory:", ["dummy"], isolation_level="IMMEDIATE")
    cursor = connection.cursor()
    cursor.execute('SELECT get_metadata("dummy://")')
    assert cursor.fetchall() == [('{"hello": "world"}',)]


def test_version_from_sql() -> None:
    """
    Test calling ``version`` from SQL.
    """
    connection = connect(":memory:")
    cursor = connection.cursor()
    cursor.execute("SELECT version()")
    shillelagh_version = distribution("shillelagh").version
    apsw_version = apsw.apswversion()  # pylint: disable=c-extension-no-member
    version = f"{shillelagh_version} (apsw {apsw_version})"
    assert cursor.fetchall() == [(version,)]


def test_date_trunc() -> None:
    """
    Test the ``date_trunc`` function.
    """
    assert date_trunc(None, "YEAR") is None

    assert date_trunc("2024-02-03T04:05:06.700000", "YEAR") == "2024-01-01T00:00:00"
    assert date_trunc("2024-02-03T04:05:06.700000", "QUARTER") == "2024-01-01T00:00:00"
    assert date_trunc("2024-02-03T04:05:06.700000", "MONTH") == "2024-02-01T00:00:00"
    assert date_trunc("2024-02-03T04:05:06.700000", "WEEK") == "2024-01-29T00:00:00"
    assert date_trunc("2024-02-03T04:05:06.700000", "DAY") == "2024-02-03T00:00:00"
    assert date_trunc("2024-02-03T04:05:06.700000", "HOUR") == "2024-02-03T04:00:00"
    assert date_trunc("2024-02-03T04:05:06.700000", "MINUTE") == "2024-02-03T04:05:00"
    assert date_trunc("2024-02-03T04:05:06.700000", "SECOND") == "2024-02-03T04:05:06"

    with pytest.raises(ValueError) as excinfo:
        date_trunc("2024-02-03T04:05:06.700000", "INVALID")
    assert str(excinfo.value) == "Unsupported truncation unit: invalid"


def test_upgrade(mocker: MockerFixture) -> None:
    """
    Test ``upgrade``.
    """
    pip = mocker.patch("shillelagh.functions.pip")
    importlib = mocker.patch("shillelagh.functions.importlib")

    output = upgrade("1.2.3")

    assert output == "Upgrade to 1.2.3 successful."
    pip.main.assert_called_with(["install", "shillelagh==1.2.3"])
    importlib.reload.assert_called_with(sys.modules["shillelagh"])


def test_upgrade_invalid_version() -> None:
    """
    Test ``upgrade`` with an invalid version.
    """
    output = upgrade("1.2.3 malicious_package=0.1")
    assert output == "Invalid version: 1.2.3 malicious_package=0.1"


def test_upgrade_exception(mocker: MockerFixture) -> None:
    """
    Test ``upgrade`` with an exception.
    """
    pip = mocker.patch("shillelagh.functions.pip")
    pip.main.side_effect = Exception("Boom!")

    output = upgrade("1.2.3")
    assert output == "Upgrade failed: Boom!"



"""
Tests for the Multicorn2 DB API 2.0 wrapper.
"""

# pylint: disable=invalid-name, redefined-outer-name, no-member, redefined-builtin

import psycopg2
import pytest
from pytest_mock import MockerFixture

from shillelagh.adapters.registry import AdapterLoader
from shillelagh.backends.multicorn.db import Cursor, CursorFactory, connect
from shillelagh.exceptions import ProgrammingError

from ...fakes import FakeAdapter


def test_connect(mocker: MockerFixture, registry: AdapterLoader) -> None:
    """
    Test the ``connect`` function.
    """
    psycopg2 = mocker.patch("shillelagh.backends.multicorn.db.psycopg2")
    CursorFactory = mocker.patch("shillelagh.backends.multicorn.db.CursorFactory")

    registry.add("dummy", FakeAdapter)

    connect(
        None,
        ["dummy"],
        user="username",
        password="password",
        host="host",
        port=1234,
        database="database",
    )
    psycopg2.connect.assert_called_with(
        None,
        cursor_factory=CursorFactory(
            {"dummy": FakeAdapter},
            {},
            "main",
        ),
        user="username",
        password="password",
        host="host",
        port=1234,
        database="database",
    )


def test_cursor_factory(mocker: MockerFixture) -> None:
    """
    Test the ``CursorFactory`` class.
    """
    Cursor = mocker.patch("shillelagh.backends.multicorn.db.Cursor")

    cursor_factory = CursorFactory(
        {"dummy": FakeAdapter},
        {},
        "main",
    )
    assert cursor_factory(
        user="username",
        password="password",
        host="host",
        port=1234,
        database="database",
    ) == Cursor(
        adapters=["dummy"],
        adapter_kwargs={},
        schema="main",
        user="username",
        password="password",
        host="host",
        port=1234,
        database="database",
    )


def test_cursor(mocker: MockerFixture) -> None:
    """
    Test the ``Cursor`` class.
    """
    mocker.patch("shillelagh.backends.multicorn.db.uuid4", return_value="uuid")
    super = mocker.patch("shillelagh.backends.multicorn.db.super", create=True)
    execute = mocker.MagicMock(name="execute")
    super.return_value.execute = execute

    cursor = Cursor(
        adapters={"dummy": FakeAdapter},
        adapter_kwargs={},
        schema="main",
    )

    cursor.execute("SELECT 1")
    execute.assert_has_calls(
        [
            mocker.call('SAVEPOINT "uuid"'),
            mocker.call("SELECT 1", None),
        ],
    )

    execute.reset_mock()
    execute.side_effect = [
        True,  # SAVEPOINT
        psycopg2.errors.UndefinedTable('relation "dummy://" does not exist'),
        True,  # ROLLBACK
        True,  # CREATE SERVER
        True,  # CREATE FOREIGN TABLE
        True,  # SAVEPOINT
        True,  # successful query
    ]

    cursor.execute('SELECT * FROM "dummy://"')
    execute.assert_has_calls(
        [
            mocker.call('SAVEPOINT "uuid"'),
            mocker.call('SELECT * FROM "dummy://"', None),
            mocker.call('ROLLBACK TO SAVEPOINT "uuid"'),
            mocker.call(
                """
CREATE SERVER shillelagh foreign data wrapper multicorn options (
    wrapper 'shillelagh.backends.multicorn.fdw.MulticornForeignDataWrapper'
);
    """,
            ),
            mocker.call(
                """
CREATE FOREIGN TABLE "dummy://" (
    "age" REAL, "name" TEXT, "pets" INTEGER
) server shillelagh options (
    adapter \'dummy\',
    args \'qQA=\'
);
        """,
            ),
            mocker.call('SAVEPOINT "uuid"'),
            mocker.call('SELECT * FROM "dummy://"', None),
        ],
    )


def test_cursor_no_table_match(mocker: MockerFixture) -> None:
    """
    Test an edge case where ``UndefinedTable`` is raised with a different message.
    """
    super = mocker.patch("shillelagh.backends.multicorn.db.super", create=True)
    execute = mocker.MagicMock(name="execute")
    super.return_value.execute = execute

    execute.side_effect = [
        True,  # SAVEPOINT
        psycopg2.errors.UndefinedTable("An unexpected error occurred"),
    ]

    cursor = Cursor(
        adapters={"dummy": FakeAdapter},
        adapter_kwargs={},
        schema="main",
    )

    with pytest.raises(ProgrammingError) as excinfo:
        cursor.execute('SELECT * FROM "dummy://"')
    assert str(excinfo.value) == "An unexpected error occurred"


def test_cursor_no_table_name(mocker: MockerFixture) -> None:
    """
    Test an edge case where we can't determine the table name from the exception.
    """
    super = mocker.patch("shillelagh.backends.multicorn.db.super", create=True)
    execute = mocker.MagicMock(name="execute")
    super.return_value.execute = execute

    execute.side_effect = [
        True,  # SAVEPOINT
        psycopg2.errors.UndefinedTable('relation "invalid://" does not exist'),
    ]

    cursor = Cursor(
        adapters={"dummy": FakeAdapter},
        adapter_kwargs={},
        schema="main",
    )

    with pytest.raises(ProgrammingError) as excinfo:
        cursor.execute('SELECT * FROM "dummy://"')
    assert str(excinfo.value) == "Could not determine table name"


def test_drop_table(mocker: MockerFixture) -> None:
    """
    Test the ``drop_table`` method.
    """
    super = mocker.patch("shillelagh.backends.multicorn.db.super", create=True)
    execute = mocker.MagicMock(name="execute")
    super.return_value.execute = execute
    adapter = mocker.MagicMock(name="adapter")
    mocker.patch(
        "shillelagh.backends.multicorn.db.find_adapter",
        return_value=(adapter, ["one"], {"two": 2}),
    )

    cursor = Cursor(
        adapters={"dummy": FakeAdapter},
        adapter_kwargs={},
        schema="main",
    )

    cursor.execute('DROP TABLE "dummy://"')
    adapter.assert_called_with("one", two=2)
    adapter().drop_table.assert_called()


def test_table_without_columns(mocker: MockerFixture) -> None:
    """
    Test an edge case where a virtual table has no columns.
    """
    super = mocker.patch("shillelagh.backends.multicorn.db.super", create=True)
    execute = mocker.MagicMock(name="execute")
    super.return_value.execute = execute
    adapter = mocker.MagicMock(name="adapter")
    adapter().get_columns.return_value = []
    mocker.patch(
        "shillelagh.backends.multicorn.db.find_adapter",
        return_value=(adapter, ["one"], {"two": 2}),
    )

    cursor = Cursor(
        adapters={"dummy": adapter},
        adapter_kwargs={},
        schema="main",
    )
    with pytest.raises(ProgrammingError) as excinfo:
        cursor._create_table("dummy://")  # pylint: disable=protected-access
    assert str(excinfo.value) == "Virtual table dummy:// has no columns"



"""
Tests for shillelagh.adapters.api.gsheets.lib.
"""

import itertools
from typing import cast

import dateutil.tz
import pytest
from pytest_mock import MockerFixture

from shillelagh.adapters.api.gsheets.fields import (
    GSheetsBoolean,
    GSheetsDate,
    GSheetsDateTime,
    GSheetsNumber,
    GSheetsString,
    GSheetsTime,
)
from shillelagh.adapters.api.gsheets.lib import (
    format_error_message,
    gen_letters,
    get_credentials,
    get_field,
    get_index_from_letters,
    get_sync_mode,
    get_url,
    get_value_from_cell,
    get_values_from_row,
)
from shillelagh.adapters.api.gsheets.types import SyncMode
from shillelagh.adapters.api.gsheets.typing import QueryResultsError
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Order
from shillelagh.filters import Equal, IsNotNull, IsNull, Like, NotEqual, Range


def test_get_field() -> None:
    """
    Test ``get_field``.
    """
    assert get_field({"type": "string"}, None) == GSheetsString(
        [Range, Equal, NotEqual, Like, IsNull, IsNotNull],
        Order.ANY,
        True,
    )
    assert get_field({"type": "number"}, None) == GSheetsNumber(
        [Range, Equal, NotEqual, IsNull, IsNotNull],
        Order.ANY,
        True,
    )
    assert get_field({"type": "boolean"}, None) == GSheetsBoolean(
        [Equal, NotEqual, IsNull, IsNotNull],
        Order.ANY,
        True,
    )
    assert get_field({"type": "date"}, None) == GSheetsDate(
        [Range, Equal, NotEqual, IsNull, IsNotNull],
        Order.ANY,
        True,
    )
    assert get_field({"type": "datetime"}, None) == GSheetsDateTime(
        [Range, Equal, NotEqual, IsNull, IsNotNull],
        Order.ANY,
        True,
    )
    timezone = dateutil.tz.gettz("America/Los_Angeles")
    assert get_field(
        {"type": "datetime", "pattern": "M/d/yyyy H:mm:ss"},
        timezone,
    ) == GSheetsDateTime(
        [Range, Equal, NotEqual, IsNull, IsNotNull],
        Order.ANY,
        True,
        "M/d/yyyy H:mm:ss",
        timezone,
    )
    assert get_field({"type": "timeofday"}, None) == GSheetsTime(
        [Range, Equal, NotEqual, IsNull, IsNotNull],
        Order.ANY,
        True,
    )
    assert get_field(
        {"type": "datetime", "pattern": "h:mm:ss am/pm"},
        None,
    ) == GSheetsTime(
        [Range, Equal, NotEqual, IsNull, IsNotNull],
        Order.ANY,
        True,
        "h:mm:ss am/pm",
    )
    assert get_field({"type": "invalid"}, None) == GSheetsString(
        [Range, Equal, NotEqual, Like, IsNull, IsNotNull],
        Order.ANY,
        True,
    )
    assert get_field(
        {
            "id": "A",
            "label": "Send time",
            "type": "timeofday",
            "pattern": "M/D/YY h:mm",
        },
        timezone,
    ) == GSheetsDateTime(
        [Range, Equal, NotEqual, IsNull, IsNotNull],
        Order.ANY,
        True,
        "M/D/YY h:mm",
        timezone,
    )


def test_format_error_message() -> None:
    """
    Test ``format_error_message``.
    """
    assert format_error_message([]) == ""
    response = {
        "version": "0.6",
        "reqId": "0",
        "status": "error",
        "errors": [
            {
                "reason": "invalid_query",
                "message": "INVALID_QUERY",
                "detailed_message": "Invalid query: NO_COLUMN: C",
            },
        ],
    }
    assert (
        format_error_message(cast(list[QueryResultsError], response["errors"]))
        == "Invalid query: NO_COLUMN: C"
    )


def test_get_url() -> None:
    """
    Test ``get_url``.
    """
    assert (
        get_url(
            "https://docs.google.com/spreadsheets/d/1/edit#gid=0",
        )
        == "https://docs.google.com/spreadsheets/d/1/gviz/tq?gid=0"
    )
    assert (
        get_url(
            "https://docs.google.com/spreadsheets/d/1/edit#gid=0",
            headers=2,
            gid=3,
            sheet="some-sheet",
        )
        == "https://docs.google.com/spreadsheets/d/1/gviz/tq?headers=2&sheet=some-sheet"
    )
    assert (
        get_url(
            "https://docs.google.com/spreadsheets/d/1/edit?headers=2&gid=1",
        )
        == "https://docs.google.com/spreadsheets/d/1/gviz/tq?headers=2&gid=1"
    )
    assert (
        get_url(
            "https://docs.google.com/spreadsheets/d/1/edit?headers=2&sheet=some-sheet",
        )
        == "https://docs.google.com/spreadsheets/d/1/gviz/tq?headers=2&sheet=some-sheet"
    )


def test_get_sync_mode() -> None:
    """
    Test ``get_sync_mode``.
    """
    assert get_sync_mode("https://docs.google.com/spreadsheets/d/1/edit#gid=42") is None
    assert (
        get_sync_mode(
            "https://docs.google.com/spreadsheets/d/1/edit?sync_mode=BATCH#gid=42",
        )
        == SyncMode.BATCH
    )
    assert (
        get_sync_mode(
            "https://docs.google.com/spreadsheets/d/1/edit?sync_mode=batch#gid=42",
        )
        == SyncMode.BATCH
    )
    assert (
        get_sync_mode(
            "https://docs.google.com/spreadsheets/d/1/edit?sync_mode=1#gid=42",
        )
        == SyncMode.BIDIRECTIONAL
    )
    with pytest.raises(ProgrammingError) as excinfo:
        get_sync_mode(
            "https://docs.google.com/spreadsheets/d/1/edit?sync_mode=0#gid=42",
        )
    assert str(excinfo.value) == "Invalid sync mode: 0"
    with pytest.raises(ProgrammingError) as excinfo:
        get_sync_mode(
            "https://docs.google.com/spreadsheets/d/1/edit?sync_mode=INVALID#gid=42",
        )
    assert str(excinfo.value) == "Invalid sync mode: INVALID"


def test_gen_letters() -> None:
    """
    Test ``gen_letters``.
    """
    letters = list(itertools.islice(gen_letters(), 60))
    assert letters == [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
        "K",
        "L",
        "M",
        "N",
        "O",
        "P",
        "Q",
        "R",
        "S",
        "T",
        "U",
        "V",
        "W",
        "X",
        "Y",
        "Z",
        "AA",
        "AB",
        "AC",
        "AD",
        "AE",
        "AF",
        "AG",
        "AH",
        "AI",
        "AJ",
        "AK",
        "AL",
        "AM",
        "AN",
        "AO",
        "AP",
        "AQ",
        "AR",
        "AS",
        "AT",
        "AU",
        "AV",
        "AW",
        "AX",
        "AY",
        "AZ",
        "AAA",
        "AAB",
        "AAC",
        "AAD",
        "AAE",
        "AAF",
        "AAG",
        "AAH",
    ]


def test_get_index_from_letters() -> None:
    """
    Test ``get_index_from_letters``.
    """
    assert get_index_from_letters("A") == 0
    assert get_index_from_letters("Z") == 25
    assert get_index_from_letters("AA") == 26
    assert get_index_from_letters("AB") == 27
    assert get_index_from_letters("AAA") == 702


def test_get_values_from_row() -> None:
    """
    Test ``get_values_from_row``.
    """
    column_map = {"country": "Z", "cnt": "AB"}
    row = {"country": "BR", "cnt": 10}
    assert get_values_from_row(row, column_map) == 25 * [""] + ["BR", "", 10]  # type: ignore


def test_get_credentials(mocker: MockerFixture):
    """
    Test ``get_credentials``.
    """
    service_account = mocker.patch(
        "shillelagh.adapters.api.gsheets.lib.google.oauth2.service_account.Credentials",
    )
    credentials = mocker.patch(
        "shillelagh.adapters.api.gsheets.lib.google.oauth2.credentials.Credentials",
    )
    app_default_credentials = mocker.patch(
        "shillelagh.adapters.api.gsheets.lib.google.auth.default",
    )

    # no credentials
    get_credentials(None, None, None, None)
    credentials.assert_not_called()
    service_account.assert_not_called()
    app_default_credentials.assert_not_called()

    # access_token
    get_credentials("token", None, None, None)
    credentials.assert_called_with("token")
    credentials.reset_mock()
    service_account.assert_not_called()
    app_default_credentials.assert_not_called()

    # service_account_file
    get_credentials(None, "credentials.json", None, None)
    credentials.assert_not_called()
    service_account.from_service_account_file.assert_called_with(
        "credentials.json",
        scopes=[
            "https://www.googleapis.com/auth/drive.readonly",
            "https://www.googleapis.com/auth/spreadsheets",
            "https://spreadsheets.google.com/feeds",
        ],
        subject=None,
    )
    app_default_credentials.assert_not_called()
    service_account.reset_mock()

    # service_account_info
    get_credentials(None, None, {"secret": "XXX"}, "user@example.com")
    credentials.assert_not_called()
    service_account.from_service_account_info.assert_called_with(
        {"secret": "XXX"},
        scopes=[
            "https://www.googleapis.com/auth/drive.readonly",
            "https://www.googleapis.com/auth/spreadsheets",
            "https://spreadsheets.google.com/feeds",
        ],
        subject="user@example.com",
    )
    app_default_credentials.assert_not_called()
    service_account.reset_mock()

    # app_default_credentials
    get_credentials(app_default_credentials=True)
    credentials.assert_not_called()
    service_account.assert_not_called()
    app_default_credentials.assert_called()


def test_get_value_from_cell() -> None:
    """
    Test ``get_value_from_cell``.
    """
    assert (
        get_value_from_cell({"v": "Date(2018,8,1,0,0,0)", "f": "9/1/2018 0:00:00"})
        == "9/1/2018 0:00:00"
    )
    assert get_value_from_cell({"v": "test"}) == "test"
    assert get_value_from_cell({"v": 1.0, "f": "1"}) == "1"
    assert get_value_from_cell({"v": True, "f": "TRUE"}) == "TRUE"
    assert get_value_from_cell(None) == ""
    assert get_value_from_cell({"v": None}) == ""



# pylint: disable=invalid-name
"""
Tests for shillelagh.adapters.api.gsheets.fields.
"""

import datetime

import dateutil.tz

from shillelagh.adapters.api.gsheets.fields import (
    GSheetsBoolean,
    GSheetsDate,
    GSheetsDateTime,
    GSheetsDuration,
    GSheetsNumber,
    GSheetsString,
    GSheetsTime,
)
from shillelagh.fields import ISODateTime, Order


def test_comparison() -> None:
    """
    Test that a GSheets field is different from a standard field.
    """
    assert GSheetsDateTime([], Order.NONE, True) != ISODateTime([], Order.NONE, True)
    assert GSheetsDateTime([], Order.NONE, True) == GSheetsDateTime(
        [],
        Order.NONE,
        True,
    )


def test_GSheetsDateTime() -> None:
    """
    Test ``GSheetsDateTime``.
    """
    assert GSheetsDateTime().parse(None) is None
    assert GSheetsDateTime().parse("") is None

    assert GSheetsDateTime(pattern="M/d/yyyy H:mm:ss").parse(
        "12/31/2020 12:34:56",
    ) == datetime.datetime(
        2020,
        12,
        31,
        12,
        34,
        56,
    )

    assert GSheetsDateTime().format(None) == ""
    assert (
        GSheetsDateTime(pattern="M/d/yyyy H:mm:ss").format(
            datetime.datetime(2020, 12, 31, 12, 34, 56),
        )
        == "12/31/2020 12:34:56"
    )

    assert GSheetsDateTime().quote(None) == "null"
    assert GSheetsDateTime().quote("") == "null"
    assert (
        GSheetsDateTime(pattern="M/d/yyyy H:mm:ss").quote("12/31/2020 12:34:56")
        == "datetime '2020-12-31 12:34:56'"
    )


def test_GSheetsDateTime_timezone() -> None:
    """
    Test GSheetsDateTime when timezone is set.
    """
    timezone = dateutil.tz.gettz("America/Los_Angeles")

    assert GSheetsDateTime(pattern="M/d/yyyy H:mm:ss", timezone=timezone).parse(
        "12/31/2020 12:34:56",
    ) == datetime.datetime(2020, 12, 31, 12, 34, 56, tzinfo=timezone)

    assert (
        GSheetsDateTime(pattern="M/d/yyyy H:mm:ss", timezone=timezone).format(
            datetime.datetime(2020, 12, 31, 12, 34, 56, tzinfo=datetime.timezone.utc),
        )
        == "12/31/2020 4:34:56"
    )


def test_GSheetsDate() -> None:
    """
    Test ``GSheetsDate``.
    """
    assert GSheetsDate().parse(None) is None
    assert GSheetsDate().parse("") is None

    assert GSheetsDate(pattern="M/d/yyyy").parse("12/31/2020") == datetime.date(
        2020,
        12,
        31,
    )

    assert GSheetsDate().format(None) == ""
    assert (
        GSheetsDate(pattern="M/d/yyyy").format(datetime.date(2020, 12, 31))
        == "12/31/2020"
    )

    assert GSheetsDate().quote(None) == "null"
    assert GSheetsDate().quote("") == "null"
    assert GSheetsDate(pattern="M/d/yyyy").quote("12/31/2020") == "date '2020-12-31'"


def test_GSheetsTime() -> None:
    """
    Test ``GSheetsTime``.
    """
    assert GSheetsTime().parse(None) is None
    assert GSheetsTime().parse("") is None

    assert GSheetsTime(pattern="h:mm:ss am/pm").parse("12:34:56 AM") == datetime.time(
        0,
        34,
        56,
    )

    assert GSheetsTime().format(None) == ""
    assert (
        GSheetsTime(pattern="h:mm:ss am/pm").format(datetime.time(12, 34, 56))
        == "12:34:56 PM"
    )

    assert GSheetsTime().quote(None) == "null"
    assert GSheetsTime().quote("") == "null"
    assert (
        GSheetsTime(pattern="h:mm:ss am/pm").quote("12:34:56 AM")
        == "timeofday '00:34:56'"
    )


def test_GSheetsDuration() -> None:
    """
    Test ``GSheetsDuration``.
    """
    assert GSheetsDuration().parse(None) is None
    assert GSheetsDuration().parse("") is None

    assert GSheetsDuration(pattern="[h]:mm:ss").parse("12:34:56") == datetime.timedelta(
        hours=12,
        minutes=34,
        seconds=56,
    )

    assert GSheetsDuration().format(None) == ""
    assert (
        GSheetsDuration(pattern="[h]:mm:ss").format(
            datetime.timedelta(hours=12, minutes=34, seconds=56),
        )
        == "12:34:56"
    )

    assert GSheetsDuration().quote(None) == "null"
    assert GSheetsDuration().quote("") == "null"
    assert (
        GSheetsDuration(pattern="[h]:mm:ss").quote("12:34:56")
        == "datetime '1899-12-30 12:34:56'"
    )


def test_GSheetsBoolean() -> None:
    """
    Test ``GSheetsBoolean``.
    """
    assert GSheetsBoolean().parse(None) is None
    assert GSheetsBoolean().parse("") is None
    assert GSheetsBoolean().parse("TRUE") is True
    assert GSheetsBoolean().parse("FALSE") is False

    assert GSheetsBoolean().format(None) == ""
    assert GSheetsBoolean().format(True) == "TRUE"
    assert GSheetsBoolean().format(False) == "FALSE"

    assert GSheetsBoolean().quote(None) == "null"
    assert GSheetsBoolean().quote("") == "null"
    assert GSheetsBoolean().quote("TRUE") == "true"
    assert GSheetsBoolean().quote("FALSE") == "false"


def test_GSheetsNumber() -> None:
    """
    Test ``GSheetsNumber``.
    """
    assert GSheetsNumber().parse(None) is None
    assert GSheetsNumber().parse("") is None
    assert GSheetsNumber().parse("1") == 1.0
    assert GSheetsNumber().parse("1.0") == 1.0
    assert GSheetsNumber(pattern="# ???/???").parse("5   1/8  ") == 5.125

    assert isinstance(GSheetsNumber().parse("1"), int)
    assert isinstance(GSheetsNumber().parse("1.0"), float)

    assert GSheetsNumber().format(None) == ""
    assert GSheetsNumber().format(1) == "1"
    assert GSheetsNumber().format(1.0) == "1.0"
    assert GSheetsNumber(pattern="# ???/???").format(5.125) == "5   1/8  "

    assert GSheetsNumber().quote(None) == "null"
    assert GSheetsNumber().quote("") == "null"
    assert GSheetsNumber().quote(1) == "1"
    assert GSheetsNumber().quote(1.0) == "1.0"


def test_GSheetsString() -> None:
    """
    Test ``GSheetsString``.
    """
    assert GSheetsString().parse(None) is None
    assert GSheetsString().parse("") is None
    assert GSheetsString().parse("test") == "test"

    assert GSheetsString().format(None) == ""
    assert GSheetsString().format("test") == "test"

    assert GSheetsString().quote(None) == "null"
    assert GSheetsString().quote("") == "''"
    assert GSheetsString().quote("test") == "'test'"



# pylint: disable=too-many-lines
"""
Tests for the Datasette adapter.
"""

from datetime import datetime, timezone

import pytest
from freezegun import freeze_time
from pytest_mock import MockerFixture
from requests import Session
from requests_mock.mocker import Mocker

from shillelagh.adapters.api.github import GitHubAPI, participation_processor
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError
from shillelagh.filters import Equal

from ...fakes import (
    github_issues_response,
    github_pulls_response,
    github_single_response,
)


def test_github(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test a simple request.
    """
    mocker.patch(
        "shillelagh.adapters.api.github.get_session",
        return_value=Session(),
    )

    page1_url = "https://api.github.com/repos/apache/superset/pulls?state=all&per_page=100&page=1"
    requests_mock.get(page1_url, json=github_pulls_response)
    page2_url = "https://api.github.com/repos/apache/superset/pulls?state=all&per_page=100&page=2"
    requests_mock.get(page2_url, json=[])

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = """
        SELECT * FROM
        "https://api.github.com/repos/apache/superset/pulls"
    """
    data = list(cursor.execute(sql))
    assert data == [
        (
            "https://github.com/apache/superset/pull/16581",
            726927278,
            16581,
            "open",
            "feat: Arash/datasets and reports",
            48933336,
            "AAfghahi",
            False,
            "arash/datasetsAndReports",
            datetime(2021, 9, 3, 15, 57, 37, tzinfo=timezone.utc),
            datetime(2021, 9, 3, 15, 57, 39, tzinfo=timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16576",
            726586317,
            16576,
            "open",
            "chore(viz): remove legacy table viz code",
            33317356,
            "villebro",
            True,
            "villebro/remove-table-viz",
            datetime(2021, 9, 3, 7, 52, 18, tzinfo=timezone.utc),
            datetime(2021, 9, 3, 8, 48, 45, tzinfo=timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16571",
            726278427,
            16571,
            "open",
            "chore(deps-dev): bump storybook-addon-jsx from 7.3.3 to 7.3.13 in /superset-frontend",
            49699333,
            "dependabot[bot]",
            False,
            "dependabot/npm_and_yarn/superset-frontend/storybook-addon-jsx-7.3.13",
            datetime(2021, 9, 2, 20, 51, 50, tzinfo=timezone.utc),
            datetime(2021, 9, 2, 21, 22, 54, tzinfo=timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16569",
            726107410,
            16569,
            "open",
            "docs: versioned _export Stable",
            47772523,
            "amitmiran137",
            False,
            "version_export_ff_on",
            datetime(2021, 9, 2, 16, 52, 34, tzinfo=timezone.utc),
            datetime(2021, 9, 2, 18, 6, 27, tzinfo=timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16566",
            725808286,
            16566,
            "open",
            "fix(docker): add ecpg to docker image",
            33317356,
            "villebro",
            False,
            "villebro/libecpg",
            datetime(2021, 9, 2, 12, 1, 2, tzinfo=timezone.utc),
            datetime(2021, 9, 2, 12, 6, 50, tzinfo=timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16564",
            725669631,
            16564,
            "open",
            "refactor: orderby control refactoring",
            2016594,
            "zhaoyongjie",
            True,
            "refactor_orderby",
            datetime(2021, 9, 2, 9, 45, 40, tzinfo=timezone.utc),
            datetime(2021, 9, 3, 10, 31, 4, tzinfo=timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16554",
            724863880,
            16554,
            "open",
            "refactor: Update async query init to support runtime feature flags",
            296227,
            "robdiciuccio",
            False,
            "rd/async-query-init-refactor",
            datetime(2021, 9, 1, 19, 51, 51, tzinfo=timezone.utc),
            datetime(2021, 9, 1, 22, 29, 46, tzinfo=timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16549",
            724669525,
            16549,
            "open",
            "feat(dashboard): Native filters - add type to native filter configuration",
            12539911,
            "m-ajay",
            False,
            "feat/migration-add-type-to-native-filter",
            datetime(2021, 9, 1, 16, 35, 50, tzinfo=timezone.utc),
            datetime(2021, 9, 3, 17, 33, 42, tzinfo=timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16548",
            724668038,
            16548,
            "open",
            "refactor: sql_json view endpoint: encapsulate ctas parameters",
            35701650,
            "ofekisr",
            False,
            "refactor/sql_json_view4",
            datetime(2021, 9, 1, 16, 33, 45, tzinfo=timezone.utc),
            datetime(2021, 9, 1, 17, 6, 32, tzinfo=timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16545",
            724509555,
            16545,
            "open",
            "perf(dashboard): decrease number of rerenders of FiltersBadge",
            15073128,
            "kgabryje",
            False,
            "perf/dashboard-rerenders-4",
            datetime(2021, 9, 1, 13, 41, 12, tzinfo=timezone.utc),
            datetime(2021, 9, 2, 15, 39, 15, tzinfo=timezone.utc),
            None,
            None,
        ),
    ]


def test_github_single_resource(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test a request to a single resource.
    """
    mocker.patch(
        "shillelagh.adapters.api.github.get_session",
        return_value=Session(),
    )

    url = "https://api.github.com/repos/apache/superset/pulls/16581"
    requests_mock.get(url, json=github_single_response)

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = """
        SELECT * FROM
        "https://api.github.com/repos/apache/superset/pulls"
        WHERE number = 16581
    """
    data = list(cursor.execute(sql))
    assert data == [
        (
            "https://github.com/apache/superset/pull/16581",
            726927278,
            16581,
            "open",
            "feat: Arash/datasets and reports",
            48933336,
            "AAfghahi",
            False,
            "arash/datasetsAndReports",
            datetime(2021, 9, 3, 15, 57, 37, tzinfo=timezone.utc),
            datetime(2021, 9, 3, 15, 57, 39, tzinfo=timezone.utc),
            None,
            None,
        ),
    ]


def test_github_single_resource_with_offset(
    mocker: MockerFixture,
    requests_mock: Mocker,
) -> None:
    """
    Test a request to a single resource.
    """
    mocker.patch(
        "shillelagh.adapters.api.github.get_session",
        return_value=Session(),
    )

    url = "https://api.github.com/repos/apache/superset/pulls/16581"
    requests_mock.get(url, json=github_single_response)

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = """
        SELECT * FROM
        "https://api.github.com/repos/apache/superset/pulls"
        WHERE number = 16581
        LIMIT 1
        OFFSET 2
    """
    data = list(cursor.execute(sql))
    assert data == []


def test_github_rate_limit(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test that the adapter was rate limited by the API.
    """
    mocker.patch(
        "shillelagh.adapters.api.github.get_session",
        return_value=Session(),
    )

    url = "https://api.github.com/repos/apache/superset/pulls?state=all&per_page=100&page=1"
    requests_mock.get(
        url,
        status_code=403,
        json={
            "message": (
                "API rate limit exceeded for 66.220.13.38. (But here's the good "
                "news: Authenticated requests get a higher rate limit. Check out "
                "the documentation for more details.)"
            ),
            "documentation_url": (
                "https://docs.github.com/rest/overview/"
                "resources-in-the-rest-api#rate-limiting"
            ),
        },
    )

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = """
        SELECT * FROM
        "https://api.github.com/repos/apache/superset/pulls"
    """
    with pytest.raises(ProgrammingError) as excinfo:
        list(cursor.execute(sql))
    assert str(excinfo.value) == (
        "API rate limit exceeded for 66.220.13.38. (But here's the good "
        "news: Authenticated requests get a higher rate limit. Check out "
        "the documentation for more details.)"
    )


def test_github_auth_token(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test a simple request.
    """
    mocker.patch(
        "shillelagh.adapters.api.github.get_session",
        return_value=Session(),
    )

    single_resource_url = "https://api.github.com/repos/apache/superset/pulls/16581"
    single_resource_mock = requests_mock.get(
        single_resource_url,
        json=github_single_response,
    )
    multiple_resources_url = (
        "https://api.github.com/repos/apache/superset/pulls?"
        "state=all&per_page=100&page=1"
    )
    multiple_resources_mock = requests_mock.get(multiple_resources_url, json=[])

    connection = connect(
        ":memory:",
        adapter_kwargs={"githubapi": {"access_token": "XXX"}},
    )
    cursor = connection.cursor()

    sql = """
        SELECT * FROM
        "https://api.github.com/repos/apache/superset/pulls"
        WHERE number = 16581
    """
    list(cursor.execute(sql))
    assert single_resource_mock.last_request.headers["Authorization"] == "Bearer XXX"

    sql = """
        SELECT * FROM
        "https://api.github.com/repos/apache/superset/pulls"
    """
    list(cursor.execute(sql))
    assert multiple_resources_mock.last_request.headers["Authorization"] == "Bearer XXX"


def test_get_multiple_resources(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Tests for ``_get_multiple_resources``.
    """
    mocker.patch("shillelagh.adapters.api.github.PAGE_SIZE", new=5)
    mocker.patch(
        "shillelagh.adapters.api.github.get_session",
        return_value=Session(),
    )

    page1_url = (
        "https://api.github.com/repos/apache/superset/pulls?state=all&per_page=5&page=1"
    )
    requests_mock.get(page1_url, json=github_pulls_response[:5])
    page2_url = (
        "https://api.github.com/repos/apache/superset/pulls?state=all&per_page=5&page=2"
    )
    requests_mock.get(page2_url, json=github_pulls_response[5:])
    page3_url = (
        "https://api.github.com/repos/apache/superset/pulls?state=all&per_page=5&page=3"
    )
    requests_mock.get(page3_url, json=[])

    adapter = GitHubAPI("repos", "apache", "superset", "pulls")
    rows = adapter._get_multiple_resources(  # pylint: disable=protected-access
        {"state": Equal("all")},
    )
    assert list(rows) == [
        {
            "url": "https://github.com/apache/superset/pull/16581",
            "id": 726927278,
            "number": 16581,
            "state": "open",
            "title": "feat: Arash/datasets and reports",
            "userid": 48933336,
            "username": "AAfghahi",
            "draft": False,
            "head": "arash/datasetsAndReports",
            "created_at": "2021-09-03T15:57:37Z",
            "updated_at": "2021-09-03T15:57:39Z",
            "closed_at": None,
            "merged_at": None,
            "rowid": 0,
        },
        {
            "url": "https://github.com/apache/superset/pull/16576",
            "id": 726586317,
            "number": 16576,
            "state": "open",
            "title": "chore(viz): remove legacy table viz code",
            "userid": 33317356,
            "username": "villebro",
            "draft": True,
            "head": "villebro/remove-table-viz",
            "created_at": "2021-09-03T07:52:18Z",
            "updated_at": "2021-09-03T08:48:45Z",
            "closed_at": None,
            "merged_at": None,
            "rowid": 1,
        },
        {
            "url": "https://github.com/apache/superset/pull/16571",
            "id": 726278427,
            "number": 16571,
            "state": "open",
            "title": (
                "chore(deps-dev): bump storybook-addon-jsx from 7.3.3 to 7.3.13 in "
                "/superset-frontend"
            ),
            "userid": 49699333,
            "username": "dependabot[bot]",
            "draft": False,
            "head": "dependabot/npm_and_yarn/superset-frontend/storybook-addon-jsx-7.3.13",
            "created_at": "2021-09-02T20:51:50Z",
            "updated_at": "2021-09-02T21:22:54Z",
            "closed_at": None,
            "merged_at": None,
            "rowid": 2,
        },
        {
            "url": "https://github.com/apache/superset/pull/16569",
            "id": 726107410,
            "number": 16569,
            "state": "open",
            "title": "docs: versioned _export Stable",
            "userid": 47772523,
            "username": "amitmiran137",
            "draft": False,
            "head": "version_export_ff_on",
            "created_at": "2021-09-02T16:52:34Z",
            "updated_at": "2021-09-02T18:06:27Z",
            "closed_at": None,
            "merged_at": None,
            "rowid": 3,
        },
        {
            "url": "https://github.com/apache/superset/pull/16566",
            "id": 725808286,
            "number": 16566,
            "state": "open",
            "title": "fix(docker): add ecpg to docker image",
            "userid": 33317356,
            "username": "villebro",
            "draft": False,
            "head": "villebro/libecpg",
            "created_at": "2021-09-02T12:01:02Z",
            "updated_at": "2021-09-02T12:06:50Z",
            "closed_at": None,
            "merged_at": None,
            "rowid": 4,
        },
        {
            "url": "https://github.com/apache/superset/pull/16564",
            "id": 725669631,
            "number": 16564,
            "state": "open",
            "title": "refactor: orderby control refactoring",
            "userid": 2016594,
            "username": "zhaoyongjie",
            "draft": True,
            "head": "refactor_orderby",
            "created_at": "2021-09-02T09:45:40Z",
            "updated_at": "2021-09-03T10:31:04Z",
            "closed_at": None,
            "merged_at": None,
            "rowid": 5,
        },
        {
            "url": "https://github.com/apache/superset/pull/16554",
            "id": 724863880,
            "number": 16554,
            "state": "open",
            "title": "refactor: Update async query init to support runtime feature flags",
            "userid": 296227,
            "username": "robdiciuccio",
            "draft": False,
            "head": "rd/async-query-init-refactor",
            "created_at": "2021-09-01T19:51:51Z",
            "updated_at": "2021-09-01T22:29:46Z",
            "closed_at": None,
            "merged_at": None,
            "rowid": 6,
        },
        {
            "url": "https://github.com/apache/superset/pull/16549",
            "id": 724669525,
            "number": 16549,
            "state": "open",
            "title": "feat(dashboard): Native filters - add type to native filter configuration",
            "userid": 12539911,
            "username": "m-ajay",
            "draft": False,
            "head": "feat/migration-add-type-to-native-filter",
            "created_at": "2021-09-01T16:35:50Z",
            "updated_at": "2021-09-03T17:33:42Z",
            "closed_at": None,
            "merged_at": None,
            "rowid": 7,
        },
        {
            "url": "https://github.com/apache/superset/pull/16548",
            "id": 724668038,
            "number": 16548,
            "state": "open",
            "title": "refactor: sql_json view endpoint: encapsulate ctas parameters",
            "userid": 35701650,
            "username": "ofekisr",
            "draft": False,
            "head": "refactor/sql_json_view4",
            "created_at": "2021-09-01T16:33:45Z",
            "updated_at": "2021-09-01T17:06:32Z",
            "closed_at": None,
            "merged_at": None,
            "rowid": 8,
        },
        {
            "url": "https://github.com/apache/superset/pull/16545",
            "id": 724509555,
            "number": 16545,
            "state": "open",
            "title": "perf(dashboard): decrease number of rerenders of FiltersBadge",
            "userid": 15073128,
            "username": "kgabryje",
            "draft": False,
            "head": "perf/dashboard-rerenders-4",
            "created_at": "2021-09-01T13:41:12Z",
            "updated_at": "2021-09-02T15:39:15Z",
            "closed_at": None,
            "merged_at": None,
            "rowid": 9,
        },
    ]


def test_github_missing_field(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test a request when the response is missing a field.

    For example, some issues don't have the ``draft`` field in the response.
    """
    mocker.patch(
        "shillelagh.adapters.api.github.get_session",
        return_value=Session(),
    )

    page1_url = "https://api.github.com/repos/apache/superset/issues?state=all&per_page=100&page=1"
    requests_mock.get(page1_url, json=github_issues_response)
    page2_url = "https://api.github.com/repos/apache/superset/issues?state=all&per_page=100&page=2"
    requests_mock.get(page2_url, json=[])

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = """
        SELECT draft FROM
        "https://api.github.com/repos/apache/superset/issues"
        LIMIT 10
    """
    data = list(cursor.execute(sql))
    assert data == [
        (False,),
        (False,),
        (None,),
        (None,),
        (False,),
        (None,),
        (False,),
        (None,),
        (False,),
        (False,),
    ]


def test_github_json_field(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test a request when the response has a JSON field.
    """
    mocker.patch(
        "shillelagh.adapters.api.github.get_session",
        return_value=Session(),
    )

    page1_url = "https://api.github.com/repos/apache/superset/issues?state=all&per_page=100&page=1"
    requests_mock.get(page1_url, json=github_issues_response)
    page2_url = "https://api.github.com/repos/apache/superset/issues?state=all&per_page=100&page=2"
    requests_mock.get(page2_url, json=[])

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = """
        SELECT labels FROM
        "https://api.github.com/repos/apache/superset/issues"
        WHERE labels != '[]'
        LIMIT 10
    """
    data = list(cursor.execute(sql))
    assert data == [
        ('["size/M", "dependencies:npm", "github_actions", "packages"]',),
        ('["size/S"]',),
        ('["size/M"]',),
        ('["size/M", "api"]',),
        ('["size/L", "api"]',),
        ('["size/XS"]',),
        ('["size/XS", "dependencies:npm"]',),
        ('["size/S"]',),
        ('["size/XS", "hold:review-after-release"]',),
        ('["size/M", "review-checkpoint", "plugins"]',),
    ]


@freeze_time("2024-01-01T00:00:00Z")
def test_participation_processor() -> None:
    """
    Test the ``participation_processor`` function.
    """
    payload = {
        "all": [27, 34, 27, 31, 24],
        "owner": [0, 0, 0, 0, 0],
    }
    processed = participation_processor(payload)

    assert processed == [
        {
            "start_at": datetime(2023, 11, 27, 0, 0, tzinfo=timezone.utc),
            "end_at": datetime(2023, 12, 4, 0, 0, tzinfo=timezone.utc),
            "all": 27,
            "owner": 0,
        },
        {
            "start_at": datetime(2023, 12, 4, 0, 0, tzinfo=timezone.utc),
            "end_at": datetime(2023, 12, 11, 0, 0, tzinfo=timezone.utc),
            "all": 34,
            "owner": 0,
        },
        {
            "start_at": datetime(2023, 12, 11, 0, 0, tzinfo=timezone.utc),
            "end_at": datetime(2023, 12, 18, 0, 0, tzinfo=timezone.utc),
            "all": 27,
            "owner": 0,
        },
        {
            "start_at": datetime(2023, 12, 18, 0, 0, tzinfo=timezone.utc),
            "end_at": datetime(2023, 12, 25, 0, 0, tzinfo=timezone.utc),
            "all": 31,
            "owner": 0,
        },
        {
            "start_at": datetime(2023, 12, 25, 0, 0, tzinfo=timezone.utc),
            "end_at": datetime(2024, 1, 1, 0, 0, tzinfo=timezone.utc),
            "all": 24,
            "owner": 0,
        },
    ]


@freeze_time("2024-01-01T00:00:00Z")
def test_github_participation(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test a request to the participation stats.
    """
    mocker.patch(
        "shillelagh.adapters.api.github.get_session",
        return_value=Session(),
    )

    url = "https://api.github.com/repos/apache/superset/stats/participation"
    requests_mock.get(
        url,
        json={
            "all": [27, 34, 27, 31, 24],
            "owner": [0, 0, 0, 0, 0],
        },
    )

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = """
        SELECT * FROM
        "https://api.github.com/repos/apache/superset/stats/participation"
    """
    data = list(cursor.execute(sql))
    assert data == [
        (
            datetime(2023, 11, 27, 0, 0, tzinfo=timezone.utc),
            datetime(2023, 12, 4, 0, 0, tzinfo=timezone.utc),
            27,
            0,
        ),
        (
            datetime(2023, 12, 4, 0, 0, tzinfo=timezone.utc),
            datetime(2023, 12, 11, 0, 0, tzinfo=timezone.utc),
            34,
            0,
        ),
        (
            datetime(2023, 12, 11, 0, 0, tzinfo=timezone.utc),
            datetime(2023, 12, 18, 0, 0, tzinfo=timezone.utc),
            27,
            0,
        ),
        (
            datetime(2023, 12, 18, 0, 0, tzinfo=timezone.utc),
            datetime(2023, 12, 25, 0, 0, tzinfo=timezone.utc),
            31,
            0,
        ),
        (
            datetime(2023, 12, 25, 0, 0, tzinfo=timezone.utc),
            datetime(2024, 1, 1, 0, 0, tzinfo=timezone.utc),
            24,
            0,
        ),
    ]



"""
An adapter that scrapes HTML tables.
"""

# pylint: disable=invalid-name

import urllib.parse
from collections.abc import Iterator
from typing import Any, Optional

import pandas as pd

from shillelagh.adapters.base import Adapter
from shillelagh.adapters.memory.pandas import get_columns_from_df, get_df_data
from shillelagh.fields import Field
from shillelagh.filters import Filter
from shillelagh.lib import SimpleCostModel
from shillelagh.typing import RequestedOrder, Row

SUPPORTED_PROTOCOLS = {"http", "https", "ftp", "file"}
AVERAGE_NUMBER_OF_ROWS = 100


class HTMLTableAPI(Adapter):
    """
    An adapter for scraping HTML tables.
    """

    safe = True

    supports_limit = True
    supports_offset = True

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        parsed = urllib.parse.urlparse(uri)
        if parsed.scheme not in SUPPORTED_PROTOCOLS:
            return False
        if fast:
            return None

        # table index can be specified as an anchor; remove
        uri = urllib.parse.urlunparse(parsed._replace(fragment=""))
        try:
            dataframes = pd.read_html(uri, flavor="bs4")
        except Exception:  # pylint: disable=broad-except
            return False

        return len(dataframes) > 0

    @staticmethod
    def parse_uri(uri: str) -> tuple[str, int]:
        parsed = urllib.parse.urlparse(uri)

        # extract table index from anchor and remove from URI
        try:
            index = int(parsed.fragment or "0")
        except ValueError:
            index = 0
        uri = urllib.parse.urlunparse(parsed._replace(fragment=""))

        return uri, index

    def __init__(self, uri: str, index: int = 0):
        super().__init__()

        self.df = pd.read_html(uri, flavor="bs4")[index]
        self.columns = get_columns_from_df(self.df)

    def get_columns(self) -> dict[str, Field]:
        return self.columns

    get_cost = SimpleCostModel(AVERAGE_NUMBER_OF_ROWS)

    def get_data(
        self,
        bounds: dict[str, Filter],
        order: list[tuple[str, RequestedOrder]],
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        yield from get_df_data(self.df, self.columns, bounds, order, limit, offset)



"""
Test the Preset adapter.
"""

import re

import pytest
from pytest_mock import MockerFixture
from requests_mock.mocker import Mocker

from shillelagh.adapters.api.preset import PresetAPI, PresetWorkspaceAPI, get_urls
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError


def test_preset(requests_mock: Mocker) -> None:
    """
    Test a simple query.
    """
    # for datassette
    requests_mock.get(re.compile(".*-/versions.json.*"), status_code=404)

    requests_mock.post(
        "https://api.app.preset.io/v1/auth/",
        json={"payload": {"access_token": "SECRET"}},
    )
    requests_mock.get(
        "https://api.app.preset.io/v1/teams/",
        json={"payload": [{"id": 1, "name": "Team 1"}]},
    )

    connection = connect(
        ":memory:",
        adapter_kwargs={
            "presetapi": {
                "access_token": "XXX",
                "access_secret": "YYY",
                "cache_expiration": -1,
            },
        },
    )
    cursor = connection.cursor()

    sql = 'SELECT * FROM "https://api.app.preset.io/v1/teams/"'
    rows = list(cursor.execute(sql))
    assert rows == [(1, "Team 1")]


def test_preset_missing_token() -> None:
    """
    Test a simple query.
    """
    connection = connect(
        ":memory:",
        adapter_kwargs={"presetapi": {"cache_expiration": -1}},
    )
    cursor = connection.cursor()

    sql = 'SELECT * FROM "https://api.app.preset.io/v1/teams/"'
    with pytest.raises(ValueError) as exc_info:
        cursor.execute(sql)
    assert str(exc_info.value) == "access_token and access_secret must be provided"


def test_supports() -> None:
    """
    Test the ``supports`` method.
    """
    assert PresetAPI.supports("/etc/password") is False
    assert PresetAPI.supports("https://example.org/data.html") is False
    assert PresetAPI.supports("https://api.app.preset.io/v1/teams/") is True
    assert PresetAPI.supports("https://api.appxpreset.io/v1/teams/") is False
    assert PresetAPI.supports("https://api.app-sdx.preset.io/v1/teams/") is True
    assert (
        PresetAPI.supports(
            "https://abcdef01.us1a.app.preset.io/sqllab/?savedQueryId=1",
        )
        is False
    )

    assert PresetWorkspaceAPI.supports("https://api.app.preset.io/v1/teams/") is False
    assert (
        PresetWorkspaceAPI.supports(
            "https://abcdef01.us1a.app.preset.io/sqllab/?savedQueryId=1",
        )
        is True
    )
    assert (
        PresetWorkspaceAPI.supports(
            "https://abcdef01.us1a.app-sdx.preset.io/sqllab/?savedQueryId=1",
        )
        is True
    )


def test_get_urls() -> None:
    """
    Test the ``get_urls`` function.
    """
    gen = get_urls(
        "https://abcdef01.us1a.app-sdx.preset.io/api/v1/chart/",
        page_size=42,
    )

    url = next(gen)
    assert (
        url
        == "https://abcdef01.us1a.app-sdx.preset.io/api/v1/chart/?q=(page:0,page_size:42)"
    )
    url = next(gen)
    assert (
        url
        == "https://abcdef01.us1a.app-sdx.preset.io/api/v1/chart/?q=(page:1,page_size:42)"
    )


def test_get_urls_unable_to_parse() -> None:
    """
    Test the ``get_urls`` function when the URL query can't be parsed.
    """

    gen = get_urls("https://example.org/?q=(((")
    assert next(gen) == "https://example.org/?q=((("
    with pytest.raises(StopIteration):
        next(gen)


def test_get_urls_with_page_parameters() -> None:
    """
    Test the ``get_urls`` function when the URL already has page parameters.
    """

    gen = get_urls("https://example.org/?q=(page:0,page_size:42)")
    assert next(gen) == "https://example.org/?q=(page:0,page_size:42)"
    with pytest.raises(StopIteration):
        next(gen)


def test_preset_workspace(requests_mock: Mocker) -> None:
    """
    Test a simple query to a Preset workspace.
    """
    # for datassette
    requests_mock.get(re.compile(".*-/versions.json.*"), status_code=404)

    requests_mock.post(
        "https://api.app.preset.io/v1/auth/",
        json={"payload": {"access_token": "SECRET"}},
    )
    requests_mock.get(
        "https://abcdef01.us1a.app.preset.io/api/v1/chart/?q=(page:0,page_size:100)",
        json={"result": [{"id": 1, "slice_name": "Team 1"}]},
    )
    requests_mock.get(
        "https://abcdef01.us1a.app.preset.io/api/v1/chart/?q=(page:1,page_size:100)",
        json={"result": []},
    )

    connection = connect(
        ":memory:",
        adapter_kwargs={
            "presetworkspaceapi": {
                "access_token": "XXX",
                "access_secret": "YYY",
                "cache_expiration": -1,
            },
        },
    )
    cursor = connection.cursor()

    sql = 'SELECT * FROM "https://abcdef01.us1a.app.preset.io/api/v1/chart/"'
    rows = list(cursor.execute(sql))
    assert rows == [(1, "Team 1")]


def test_preset_workspace_pagination(requests_mock: Mocker) -> None:
    """
    Test pagination in a query to a Preset workspace.
    """
    # for datassette
    requests_mock.get(re.compile(".*-/versions.json.*"), status_code=404)

    requests_mock.post(
        "https://api.app.preset.io/v1/auth/",
        json={"payload": {"access_token": "SECRET"}},
    )
    requests_mock.get(
        "https://abcdef01.us1a.app.preset.io/api/v1/chart/?q=(page:0,page_size:100)",
        json={
            "result": [{"id": i + 1, "slice_name": f"Team {i+1}"} for i in range(100)],
        },
    )
    requests_mock.get(
        "https://abcdef01.us1a.app.preset.io/api/v1/chart/?q=(page:1,page_size:100)",
        json={
            "result": [
                {"id": i + 101, "slice_name": f"Team {i+101}"} for i in range(50)
            ],
        },
    )
    requests_mock.get(
        "https://abcdef01.us1a.app.preset.io/api/v1/chart/?q=(page:2,page_size:100)",
        json={"result": []},
    )

    connection = connect(
        ":memory:",
        adapter_kwargs={
            "presetworkspaceapi": {
                "access_token": "XXX",
                "access_secret": "YYY",
                "cache_expiration": -1,
            },
        },
    )
    cursor = connection.cursor()

    sql = 'SELECT * FROM "https://abcdef01.us1a.app.preset.io/api/v1/chart/" LIMIT 5 OFFSET 98'
    rows = list(cursor.execute(sql))
    assert rows == [
        (99, "Team 99"),
        (100, "Team 100"),
        (101, "Team 101"),
        (102, "Team 102"),
        (103, "Team 103"),
    ]


def test_preset_workspace_error(requests_mock: Mocker) -> None:
    """
    Test error handling when accessing a workspace API.
    """
    # for datassette
    requests_mock.get(re.compile(".*-/versions.json.*"), status_code=404)

    requests_mock.post(
        "https://api.app.preset.io/v1/auth/",
        json={"payload": {"access_token": "SECRET"}},
    )
    requests_mock.get(
        "https://abcdef01.us1a.app.preset.io/api/v1/chart/?q=(page:0,page_size:100)",
        json={
            "errors": [
                {
                    "message": "Your session has expired. Please refresh the page to sign in.",
                    "error_type": "GENERIC_BACKEND_ERROR",
                    "level": "error",
                    "extra": {
                        "issue_codes": [
                            {
                                "code": 1011,
                                "message": "Issue 1011 - Superset encountered an unexpected error.",
                            },
                        ],
                    },
                },
            ],
        },
        status_code=500,
    )

    connection = connect(
        ":memory:",
        adapter_kwargs={
            "presetworkspaceapi": {
                "access_token": "XXX",
                "access_secret": "YYY",
                "cache_expiration": -1,
            },
        },
    )
    cursor = connection.cursor()

    sql = 'SELECT * FROM "https://abcdef01.us1a.app.preset.io/api/v1/chart/"'
    with pytest.raises(ProgrammingError) as excinfo:
        cursor.execute(sql)
    assert (
        str(excinfo.value)
        == "Error: Your session has expired. Please refresh the page to sign in."
    )


def test_preset_workspace_no_urls(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test when no URLs are returned.
    """
    mocker.patch("shillelagh.adapters.api.preset.get_urls", return_value=[])

    requests_mock.post(
        "https://api.app.preset.io/v1/auth/",
        json={"payload": {"access_token": "SECRET"}},
    )

    adapter = PresetWorkspaceAPI(
        "https://abcdef01.us1a.app.preset.io/api/v1/chart/",
        access_token="XXX",
        access_secret="YYY",
        cache_expiration=-1,
    )
    assert list(adapter.get_data({}, [])) == []



# pylint: disable=invalid-name, c-extension-no-member, unused-import
"""
A DB API 2.0 wrapper for APSW.
"""

import datetime
import itertools
import logging
import re
from collections.abc import Iterator
from functools import partial, wraps
from typing import Any, Callable, Optional, TypeVar, cast

import apsw

from shillelagh import functions
from shillelagh.adapters.base import Adapter
from shillelagh.adapters.registry import registry
from shillelagh.backends.apsw.vt import VTModule, type_map
from shillelagh.exceptions import (  # nopycln: import; pylint: disable=redefined-builtin
    DatabaseError,
    DataError,
    Error,
    IntegrityError,
    InterfaceError,
    InternalError,
    NotSupportedError,
    OperationalError,
    ProgrammingError,
    Warning,
)
from shillelagh.fields import Blob, Field
from shillelagh.lib import (
    best_index_object_available,
    combine_args_kwargs,
    escape_identifier,
    find_adapter,
    serialize,
)
from shillelagh.types import (
    BINARY,
    DATETIME,
    NUMBER,
    ROWID,
    STRING,
    Binary,
    Date,
    DateFromTicks,
    Time,
    TimeFromTicks,
    Timestamp,
    TimestampFromTicks,
)
from shillelagh.typing import Description, SQLiteValidType

__all__ = [
    "DatabaseError",
    "DataError",
    "Error",
    "IntegrityError",
    "InterfaceError",
    "InternalError",
    "OperationalError",
    "BINARY",
    "DATETIME",
    "NUMBER",
    "ROWID",
    "STRING",
    "Binary",
    "Date",
    "DateFromTicks",
    "Time",
    "TimeFromTicks",
    "Timestamp",
    "TimestampFromTicks",
    "Warning",
]

apilevel = "2.0"
threadsafety = 2
paramstyle = "qmark"
sqlite_version_info = tuple(
    int(number) for number in apsw.sqlitelibversion().split(".")
)

NO_SUCH_TABLE = "SQLError: no such table: "
DEFAULT_SCHEMA = "main"

CURSOR_METHOD = TypeVar("CURSOR_METHOD", bound=Callable[..., Any])

_logger = logging.getLogger(__name__)


def check_closed(method: CURSOR_METHOD) -> CURSOR_METHOD:
    """Decorator that checks if a connection or cursor is closed."""

    @wraps(method)
    def wrapper(self: "Cursor", *args: Any, **kwargs: Any) -> Any:
        if self.closed:
            raise ProgrammingError(f"{self.__class__.__name__} already closed")
        return method(self, *args, **kwargs)

    return cast(CURSOR_METHOD, wrapper)


def check_result(method: CURSOR_METHOD) -> CURSOR_METHOD:
    """Decorator that checks if the cursor has results from ``execute``."""

    @wraps(method)
    def wrapper(self: "Cursor", *args: Any, **kwargs: Any) -> Any:
        if self._results is None:  # pylint: disable=protected-access
            raise ProgrammingError("Called before ``execute``")
        return method(self, *args, **kwargs)

    return cast(CURSOR_METHOD, wrapper)


def get_type_code(type_name: str) -> type[Field]:
    """
    Return a ``Field`` that corresponds to a type name.

    This is used to build the description of the cursor after a successful
    query.
    """
    return cast(type[Field], type_map.get(type_name, Blob))


def convert_binding(binding: Any) -> SQLiteValidType:
    """
    Convert a binding to a SQLite type.

    Eg, if the user is filtering a timestamp column we need to convert the
    ``datetime.datetime`` object in the binding to a string.
    """
    if isinstance(binding, bool):
        return int(binding)
    if isinstance(binding, (int, float, str, bytes, type(None))):
        return binding
    if isinstance(binding, (datetime.datetime, datetime.date, datetime.time)):
        return binding.isoformat()
    return str(binding)


class Cursor:  # pylint: disable=too-many-instance-attributes
    """
    Connection cursor.
    """

    def __init__(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        cursor: "apsw.Cursor",
        adapters: list[type[Adapter]],
        adapter_kwargs: dict[str, dict[str, Any]],
        isolation_level: Optional[str] = None,
        schema: str = DEFAULT_SCHEMA,
    ):
        self._cursor = cursor
        self._adapters = adapters
        self._adapter_kwargs = adapter_kwargs

        self.in_transaction = False
        self.isolation_level = isolation_level

        self.schema = schema

        # This read/write attribute specifies the number of rows to fetch at a
        # time with .fetchmany(). It defaults to 1 meaning to fetch a single
        # row at a time.
        self.arraysize = 1

        self.closed = False

        # this is updated only after a query
        self.description: Description = None

        # this is set to an iterator of rows after a successful query
        self._results: Optional[Iterator[tuple[Any, ...]]] = None
        self._rowcount = -1

        self.operation: Optional[str] = None

        # Approach from: https://github.com/rogerbinns/apsw/issues/160#issuecomment-33927297
        # pylint: disable=unused-argument
        def exectrace(
            cursor: "apsw.Cursor",
            sql: str,
            bindings: Optional[tuple[Any, ...]],
        ) -> bool:
            # In the case of an empty sequence, fall back to None,
            # meaning no rows returned.
            self.description = self._cursor.getdescription() or None
            return True

        self._cursor.setexectrace(exectrace)

    @property  # type: ignore
    @check_closed
    def rowcount(self) -> int:
        """
        Return the number of rows after a query.
        """
        try:
            results = list(self._results)  # type: ignore
        except TypeError:
            return -1

        n = len(results)
        self._results = iter(results)
        return max(0, self._rowcount) + n

    @check_closed
    def close(self) -> None:
        """
        Close the cursor.
        """
        self._cursor.close()
        self.closed = True

    @check_closed
    def execute(
        self,
        operation: str,
        parameters: Optional[tuple[Any, ...]] = None,
    ) -> "Cursor":
        """
        Execute a query using the cursor.
        """
        if not self.in_transaction and self.isolation_level:
            self._cursor.execute(f"BEGIN {self.isolation_level}")
            self.in_transaction = True

        self.description = None
        self._rowcount = -1

        # store current SQL in the cursor
        self.operation = operation

        # convert parameters (bindings) to types accepted by SQLite
        if parameters:
            parameters = tuple(convert_binding(parameter) for parameter in parameters)

        # this is where the magic happens: instead of forcing users to register
        # their virtual tables explicitly, we do it for them when they first try
        # to access them and it fails because the table doesn't exist yet
        while True:
            try:
                self._cursor.execute(operation, parameters)
                self.description = self._get_description()
                self._results = self._convert(self._cursor)
                break
            except apsw.SQLError as ex:
                message = ex.args[0]
                if not message.startswith(NO_SUCH_TABLE):
                    raise ProgrammingError(message) from ex

                # create the virtual table
                uri = message[len(NO_SUCH_TABLE) :]
                self._create_table(uri)

        if uri := self._drop_table_uri(operation):
            adapter, args, kwargs = find_adapter(
                uri,
                self._adapter_kwargs,
                self._adapters,
            )
            instance = adapter(*args, **kwargs)
            instance.drop_table()

        return self

    def _drop_table_uri(self, operation: str) -> Optional[str]:
        """
        Build a ``DROP TABLE`` regexp.
        """
        # remove comments
        operation = "\n".join(
            line for line in operation.split("\n") if not line.strip().startswith("--")
        )
        schema = re.escape(self.schema)
        regexp = re.compile(
            r"^\s*DROP\s+TABLE\s+(IF\s+EXISTS\s+)?"
            rf'({schema}\.)?(?P<uri>(.*?)|(".*?"))\s*;?\s*$',
            re.IGNORECASE,
        )
        if match := regexp.match(operation):
            return match.groupdict()["uri"].strip('"')

        return None

    def _convert(self, cursor: "apsw.Cursor") -> Iterator[tuple[Any, ...]]:
        """
        Convert row from SQLite types to native Python types.

        SQLite only supports 5 types. For booleans and time-related types
        we need to do the conversion here.
        """
        if not self.description:
            return  # pragma: no cover

        for row in cursor:
            yield tuple(
                # convert from SQLite types to native Python types
                type_map[desc[1].type]().parse(col)
                for col, desc in zip(row, self.description)
            )

    def _create_table(self, uri: str) -> None:
        """
        Create a virtual table.

        This method is called the first time a virtual table is accessed.
        """
        prefix = self.schema + "."
        if uri.startswith(prefix):
            uri = uri[len(prefix) :]

        adapter, args, kwargs = find_adapter(uri, self._adapter_kwargs, self._adapters)
        formatted_args = ", ".join(
            f"'{serialize(arg)}'"
            for arg in combine_args_kwargs(adapter, *args, **kwargs)
        )
        table_name = escape_identifier(uri)
        self._cursor.execute(
            f'CREATE VIRTUAL TABLE "{table_name}" USING {adapter.__name__}({formatted_args})',
        )

    def _get_description(self) -> Description:
        """
        Return the cursor description.

        We only return name and type, since that's what we get from APSW.
        """
        try:
            description = self._cursor.getdescription()
        except apsw.ExecutionCompleteError:
            return self.description

        return [
            (
                name,
                get_type_code(type_name),
                None,
                None,
                None,
                None,
                True,
            )
            for name, type_name in description
        ]

    @check_closed
    def executemany(
        self,
        operation: str,
        seq_of_parameters: Optional[list[tuple[Any, ...]]] = None,
    ) -> "Cursor":
        """
        Execute multiple statements.

        Currently not supported.
        """
        raise NotSupportedError(
            "``executemany`` is not supported, use ``execute`` instead",
        )

    @check_result
    @check_closed
    def fetchone(self) -> Optional[tuple[Any, ...]]:
        """
        Fetch the next row of a query result set, returning a single sequence,
        or ``None`` when no more data is available.
        """
        try:
            row = self.next()
        except StopIteration:
            return None

        self._rowcount = max(0, self._rowcount) + 1

        return row

    @check_result
    @check_closed
    def fetchmany(self, size=None) -> list[tuple[Any, ...]]:
        """
        Fetch the next set of rows of a query result, returning a sequence of
        sequences (e.g. a list of tuples). An empty sequence is returned when
        no more rows are available.
        """
        size = size or self.arraysize
        results = list(itertools.islice(self, size))

        return results

    @check_result
    @check_closed
    def fetchall(self) -> list[tuple[Any, ...]]:
        """
        Fetch all (remaining) rows of a query result, returning them as a
        sequence of sequences (e.g. a list of tuples). Note that the cursor's
        arraysize attribute can affect the performance of this operation.
        """
        results = list(self)

        return results

    @check_closed
    def setinputsizes(self, sizes: int) -> None:
        """
        Used before ``execute`` to predefine memory areas for parameters.

        Currently not supported.
        """

    @check_closed
    def setoutputsizes(self, sizes: int) -> None:
        """
        Set a column buffer size for fetches of large columns.

        Currently not supported.
        """

    @check_result
    @check_closed
    def __iter__(self) -> Iterator[tuple[Any, ...]]:
        for row in self._results:  # type: ignore
            self._rowcount = max(0, self._rowcount) + 1
            yield row

    @check_result
    @check_closed
    def __next__(self) -> tuple[Any, ...]:
        return next(self._results)  # type: ignore

    next = __next__


def apsw_version() -> str:
    """
    Custom implementation of the ``VERSION`` function.

    This function shows the backend version::

        sql> SELECT VERSION();
        VERSION()
        ----------------------
        1.0.5 (apsw 3.36.0-r1)

    """
    return f"{functions.version()} (apsw {apsw.apswversion()})"


class Connection:  # pylint: disable=too-many-instance-attributes
    """Connection."""

    def __init__(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        path: str,
        adapters: list[type[Adapter]],
        adapter_kwargs: dict[str, dict[str, Any]],
        isolation_level: Optional[str] = None,
        apsw_connection_kwargs: Optional[dict[str, Any]] = None,
        schema: str = DEFAULT_SCHEMA,
        safe: bool = False,
    ):
        # create underlying APSW connection
        apsw_connection_kwargs = apsw_connection_kwargs or {}
        self._connection = apsw.Connection(path, **apsw_connection_kwargs)
        self.isolation_level = isolation_level
        self.schema = schema
        self.safe = safe

        # register adapters
        for adapter in adapters:
            if best_index_object_available():
                self._connection.createmodule(
                    adapter.__name__,
                    VTModule(adapter),
                    use_bestindex_object=adapter.supports_requested_columns,
                )
            else:
                self._connection.createmodule(adapter.__name__, VTModule(adapter))
        self._adapters = adapters
        self._adapter_kwargs = adapter_kwargs

        # register functions
        available_functions = {
            "sleep": functions.sleep,
            "version": apsw_version,
            "get_metadata": partial(
                functions.get_metadata,
                self._adapter_kwargs,
                adapters,
            ),
            "date_trunc": functions.date_trunc,
        }
        if not safe:
            available_functions["upgrade"] = functions.upgrade

        for name, function in available_functions.items():
            self._connection.create_scalar_function(name, function)

        self.closed = False
        self.cursors: list[Cursor] = []

    @check_closed
    def close(self) -> None:
        """Close the connection now."""
        self.closed = True
        for cursor in self.cursors:
            if not cursor.closed:
                cursor.close()

    @check_closed
    def commit(self) -> None:
        """Commit any pending transaction to the database."""
        for cursor in self.cursors:
            if cursor.in_transaction:
                cursor._cursor.execute("COMMIT")  # pylint: disable=protected-access
                cursor.in_transaction = False

    @check_closed
    def rollback(self) -> None:
        """Rollback any transactions."""
        for cursor in self.cursors:
            if cursor.in_transaction:
                cursor._cursor.execute("ROLLBACK")  # pylint: disable=protected-access
                cursor.in_transaction = False

    @check_closed
    def cursor(self) -> Cursor:
        """Return a new Cursor Object using the connection."""
        cursor = Cursor(
            self._connection.cursor(),
            self._adapters,
            self._adapter_kwargs,
            self.isolation_level,
            self.schema,
        )
        self.cursors.append(cursor)

        return cursor

    @check_closed
    def execute(
        self,
        operation: str,
        parameters: Optional[tuple[Any, ...]] = None,
    ) -> Cursor:
        """
        Execute a query on a cursor.
        """
        cursor = self.cursor()
        return cursor.execute(operation, parameters)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.commit()
        self.close()


def connect(  # pylint: disable=too-many-arguments, too-many-positional-arguments
    path: str,
    adapters: Optional[list[str]] = None,
    adapter_kwargs: Optional[dict[str, dict[str, Any]]] = None,
    safe: bool = False,
    isolation_level: Optional[str] = None,
    apsw_connection_kwargs: Optional[dict[str, Any]] = None,
    schema: str = DEFAULT_SCHEMA,
) -> Connection:
    """
    Constructor for creating a connection to the database.
    """
    adapter_kwargs = adapter_kwargs or {}
    enabled_adapters = registry.load_all(adapters, safe)

    # replace entry point names with class names
    mapping = {
        name: adapter.__name__.lower() for name, adapter in enabled_adapters.items()
    }
    adapter_kwargs = {mapping[k]: v for k, v in adapter_kwargs.items() if k in mapping}

    return Connection(
        path,
        list(enabled_adapters.values()),
        adapter_kwargs,
        isolation_level,
        apsw_connection_kwargs,
        schema,
        safe,
    )



"""
Test the generic XML adapter.
"""

import re
import xml.etree.ElementTree as ET

import pytest
from requests_mock.mocker import Mocker
from yarl import URL

from shillelagh.adapters.api.generic_xml import element_to_dict
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError

baseurl = URL("https://api.congress.gov/v3/bill/118")


def test_element_to_dict() -> None:
    """
    Test XML to dict conversion.
    """
    xmlstr = """<?xml version="1.0" encoding="utf-8"?>
<root>
  <context attribute="value">introduction</context>
  <greetings>
    <greeting>hi</greeting>
    <greeting>fala aí</greeting>
    <greeting>hola</greeting>
  </greetings>
</root>
    """
    root = ET.fromstring(xmlstr)
    assert element_to_dict(root) == {
        "context": "introduction",
        "greetings": {"greeting": ["hi", "fala aí", "hola"]},
    }


def test_generic_xml(requests_mock: Mocker) -> None:
    """
    Test a simple query.
    """
    # for datassette
    requests_mock.get(re.compile(".*-/versions.json.*"), status_code=404)

    params = {
        "format": "Xml",
        "offset": 0,
        "limit": 2,
        "api_key": "SECRET",
    }
    url = (baseurl % params).with_fragment(".//bill")
    requests_mock.head(str(url), headers={"content-type": "application/xml"})
    requests_mock.get(
        str(url),
        text="""<?xml version="1.0" encoding="utf-8"?>
<api-root>
   <bills>
      <bill>
         <congress>
            118
         </congress>
         <type>
            SRES
         </type>
         <originChamber>
            Senate
         </originChamber>
         <originChamberCode>
            S
         </originChamberCode>
         <number>
            416
         </number>
         <url>
            https://api.congress.gov/v3/bill/118/sres/416?format=xml
         </url>
         <title>
            A resolution to authorize testimony and representation in United States v. Sullivan.
         </title>
         <updateDateIncludingText>
            2023-10-19T12:43:41Z
         </updateDateIncludingText>
         <latestAction>
            <actionDate>
               2023-10-18
            </actionDate>
            <text>
               Submitted in the Senate, considered, and agreed to without amendment and with a preamble by Unanimous Consent. (consideration: CR S5082-5083; text: CR S5091)
            </text>
         </latestAction>
         <updateDate>
            2023-10-19
         </updateDate>
      </bill>
      <bill>
         <congress>
            118
         </congress>
         <type>
            SRES
         </type>
         <originChamber>
            Senate
         </originChamber>
         <originChamberCode>
            S
         </originChamberCode>
         <number>
            415
         </number>
         <url>
            https://api.congress.gov/v3/bill/118/sres/415?format=xml
         </url>
         <title>
            A resolution to authorize testimony and representation in United States v. Samsel.
         </title>
         <updateDateIncludingText>
            2023-10-19T12:43:40Z
         </updateDateIncludingText>
         <latestAction>
            <actionDate>
               2023-10-18
            </actionDate>
            <text>
               Submitted in the Senate, considered, and agreed to without amendment and with a preamble by Unanimous Consent. (consideration: CR S5082-5083; text: CR S5091)
            </text>
         </latestAction>
         <updateDate>
            2023-10-19
         </updateDate>
      </bill>
   </bills>
   <pagination>
      <count>
         10503
      </count>
      <next>
         https://api.congress.gov/v3/bill/118?offset=2&amp;limit=2&amp;format=xml
      </next>
   </pagination>
   <request>
      <congress>
         118
      </congress>
      <contentType>
         application/xml
      </contentType>
      <format>
         xml
      </format>
   </request>
</api-root>""",
    )

    connection = connect(
        ":memory:",
        adapter_kwargs={"genericxmlapi": {"cache_expiration": -1}},
    )
    cursor = connection.cursor()

    sql = f'SELECT congress, type, latestAction FROM "{url}"'
    rows = list(cursor.execute(sql))
    assert rows == [
        (
            "118",
            "SRES",
            '{"actionDate": "2023-10-18", "text": "Submitted in the Senate, considered, '
            "and agreed to without amendment and with a preamble by Unanimous Consent. "
            '(consideration: CR S5082-5083; text: CR S5091)"}',
        ),
        (
            "118",
            "SRES",
            '{"actionDate": "2023-10-18", "text": "Submitted in the Senate, considered, '
            "and agreed to without amendment and with a preamble by Unanimous Consent. "
            '(consideration: CR S5082-5083; text: CR S5091)"}',
        ),
    ]

    requests_mock.get(
        str(url),
        text="Something went wrong",
        status_code=500,
    )
    with pytest.raises(ProgrammingError) as excinfo:
        list(cursor.execute(sql))
    assert str(excinfo.value) == "Error: Something went wrong"



"""
An adapter for in-memory holidays.
"""

import datetime
from collections.abc import Iterator
from typing import Any, Optional, TypedDict

from holidays import country_holidays, list_supported_countries

from shillelagh.adapters.base import Adapter
from shillelagh.fields import Date, Order, String
from shillelagh.filters import Equal, Range
from shillelagh.lib import SimpleCostModel
from shillelagh.typing import RequestedOrder, Row

# this is just a wild guess; used to estimate query cost
AVERAGE_NUMBER_OF_ROWS = 1000


class BoundsType(TypedDict, total=False):
    """
    The type of the bounds parameter.
    """

    country: Equal
    date: Range


class HolidaysMemory(Adapter):
    """
    An adapter for in-memory holidays.
    """

    safe = True

    supports_limit = False
    supports_offset = False
    supports_requested_columns = False

    country = String(filters=[Equal], order=Order.NONE, exact=True)
    name = String(filters=[], order=Order.NONE, exact=False)
    date = Date(filters=[Range], order=Order.ASCENDING, exact=False)

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        return uri == "holidays"

    @staticmethod
    def parse_uri(uri: str) -> tuple[()]:
        return ()

    get_cost = SimpleCostModel(AVERAGE_NUMBER_OF_ROWS)

    def get_data(  # type: ignore
        self,
        bounds: BoundsType,
        order: list[tuple[str, RequestedOrder]],
        **kwargs: Any,
    ) -> Iterator[Row]:
        countries = (
            [bounds["country"].value]
            if "country" in bounds
            else list_supported_countries()
        )

        default_start = default_end = datetime.date.today()
        time_range = bounds.get("date", Range(start=default_start, end=default_end))
        years = range(
            (time_range.start or default_start).year,
            (time_range.end or default_end).year + 1,
        )

        i = 0
        for country in countries:
            holidays = country_holidays(country, years=years)
            for date, name in sorted(holidays.items()):
                yield {
                    "rowid": i,
                    "country": country,
                    "name": name,
                    "date": date,
                }
                i += 1



"""
Tests for shillelagh.filters.
"""

import pytest

from shillelagh.filters import (
    Endpoint,
    Equal,
    Impossible,
    IsNotNull,
    IsNull,
    Like,
    NotEqual,
    Operator,
    Range,
    Side,
)


def test_equal() -> None:
    """
    Test ``Equal``.
    """
    operations = {(Operator.EQ, 10)}
    filter_ = Equal.build(operations)
    assert isinstance(filter_, Equal)
    assert filter_.value == 10


def test_equal_multiple_value() -> None:
    """
    Test multiple operations.
    """
    operations = {
        (Operator.EQ, 10),
        (Operator.EQ, 10),
    }
    filter_ = Equal.build(operations)
    assert isinstance(filter_, Equal)
    assert filter_.value == 10


def test_equal_check() -> None:
    """
    Test ``_check``.
    """
    operations = {
        (Operator.EQ, 10),
        (Operator.EQ, 10),
    }
    filter_ = Equal.build(operations)
    assert filter_.check(10)
    assert not filter_.check(20)


def test_equal_impossible() -> None:
    """
    Test impossible operations.
    """
    operations = {
        (Operator.EQ, 10),
        (Operator.EQ, 20),
    }
    filter_ = Equal.build(operations)
    assert isinstance(filter_, Impossible)


def test_not_equal() -> None:
    """
    Test ``NotEqual``.
    """
    operations = {(Operator.NE, 10)}
    filter_ = NotEqual.build(operations)
    assert isinstance(filter_, NotEqual)
    assert filter_.value == 10


def test_not_equal_multiple_value() -> None:
    """
    Test multiple operations.
    """
    operations = {
        (Operator.NE, 10),
        (Operator.NE, 10),
    }
    filter_ = NotEqual.build(operations)
    assert isinstance(filter_, NotEqual)
    assert filter_.value == 10


def test_not_equal_check() -> None:
    """
    Test ``_check``.
    """
    operations = {
        (Operator.NE, 10),
        (Operator.NE, 10),
    }
    filter_ = NotEqual.build(operations)
    assert filter_.check(20)
    assert not filter_.check(10)


def test_not_equal_impossible() -> None:
    """
    Test impossible operations.
    """
    operations = {
        (Operator.NE, 10),
        (Operator.NE, 20),
    }
    filter_ = NotEqual.build(operations)
    assert isinstance(filter_, Impossible)


def test_like() -> None:
    """
    Test ``Like``.
    """
    operations = {(Operator.LIKE, "%test%")}
    filter_ = Like.build(operations)
    assert isinstance(filter_, Like)
    assert filter_.value == "%test%"


def test_like_multiple_value() -> None:
    """
    Test multiple operations.
    """
    operations = {
        (Operator.LIKE, "%test%"),
        (Operator.LIKE, "%test%"),
    }
    filter_ = Like.build(operations)
    assert isinstance(filter_, Like)
    assert filter_.value == "%test%"


def test_like_check() -> None:
    """
    Test ``_check``.
    """
    operations = {
        (Operator.LIKE, "%test%"),
        (Operator.LIKE, "%test%"),
    }
    filter_ = Like.build(operations)
    assert filter_.check("this is a test")
    assert not filter_.check("this is not")


def test_like_impossible() -> None:
    """
    Test impossible operations.
    """
    operations = {
        (Operator.LIKE, "%foo%"),
        (Operator.LIKE, "%bar%"),
    }
    filter_ = Like.build(operations)
    assert isinstance(filter_, Impossible)


def test_range() -> None:
    """
    Test ``Range``.
    """
    operations = {
        (Operator.GT, 0),
        (Operator.LT, 10),
        (Operator.GT, 2),
        (Operator.LE, 4),
        (Operator.GE, 2),
    }
    filter_ = Range.build(operations)
    assert isinstance(filter_, Range)
    assert filter_.start == 2
    assert filter_.end == 4
    assert not filter_.include_start
    assert filter_.include_end
    assert str(filter_) == ">2,<=4"


def test_range_equal() -> None:
    """
    Test ``Range`` collapsed to an equality.
    """
    operations = {
        (Operator.GT, 0),
        (Operator.EQ, 3),
        (Operator.LT, 10),
        (Operator.GT, 2),
        (Operator.LE, 4),
        (Operator.GE, 2),
    }
    filter_ = Range.build(operations)
    assert isinstance(filter_, Range)
    assert filter_.start == 3
    assert filter_.end == 3
    assert filter_.include_start
    assert filter_.include_end
    assert str(filter_) == "==3"


def test_range_equal_impossible() -> None:
    """
    Test an impossible range.
    """
    operations = {
        (Operator.GT, 0),
        (Operator.LT, -1),
    }
    filter_ = Range.build(operations)
    assert isinstance(filter_, Impossible)

    operations = {
        (Operator.LT, -1),
        (Operator.GT, 0),
    }
    filter_ = Range.build(operations)
    assert isinstance(filter_, Impossible)


def test_range_include() -> None:
    """
    Test operations with different includes.
    """
    operations = {
        (Operator.GE, 2),
        (Operator.GT, 2),
        (Operator.LE, 4),
        (Operator.LT, 4),
    }
    filter_ = Range.build(operations)
    assert isinstance(filter_, Range)
    assert filter_.start == 2
    assert filter_.end == 4
    assert not filter_.include_start
    assert not filter_.include_end


def test_range_check() -> None:
    """
    Test ``_check`` in different ranges.
    """
    filter_ = Range(2, 4, False, True)
    assert not filter_.check(2)
    assert filter_.check(3)
    assert filter_.check(4)
    assert str(filter_) == ">2,<=4"

    filter_ = Range(2, None, True, True)
    assert not filter_.check(1)
    assert filter_.check(2)
    assert str(filter_) == ">=2"

    filter_ = Range(2, None, False, True)
    assert not filter_.check(2)
    assert str(filter_) == ">2"

    filter_ = Range(None, 4, True, True)
    assert filter_.check(4)
    assert not filter_.check(5)
    assert str(filter_) == "<=4"

    filter_ = Range(None, 4, True, False)
    assert not filter_.check(4)
    assert str(filter_) == "<4"


def test_range_invalid_operator() -> None:
    """
    Test that ``build`` raises an exception on invalid operators.
    """
    operations = {(-1, 0)}
    with pytest.raises(Exception) as excinfo:
        Range.build(operations)  # type: ignore

    assert str(excinfo.value) == "Invalid operator: -1"


def test_combine_ranges() -> None:
    """
    Test combining ranges.
    """
    assert (Range(0, 1, False, False) == 1) is False
    with pytest.raises(TypeError) as excinfo:
        _ = Range(0, 1, False, False) + 1
    assert str(excinfo.value) == "unsupported operand type(s) for +: 'Range' and 'int'"

    range1 = Range(1, 10, False, False)
    range2 = Range(2, 9, True, True)
    assert range1 + range2 == range2
    assert range2 + range1 == range2
    assert range1 + range1 == range1

    range3 = Range(None, 9, True, True)
    assert range1 + range3 == Range(1, 9, False, True)
    assert range3 + range1 == Range(1, 9, False, True)

    range4 = Range(2, None, True, True)
    assert range1 + range4 == Range(2, 10, True, False)
    assert range4 + range1 == Range(2, 10, True, False)

    range5 = Range(None, -10, True, True)
    range6 = Range(10, None, True, True)
    assert range5 + range6 == Impossible()


def test_impossible() -> None:
    """
    Test ``Impossible``.
    """
    assert Impossible.build([]) == Impossible()  # type: ignore
    assert Impossible().check(10) is False
    assert Impossible() != 0


def build_endpoint(representation: str) -> Endpoint:
    """
    Function to build an endpoint.
    """
    if representation[0] in {"(", "["}:
        return Endpoint(
            value=int(representation[1:]),
            include=representation[0] == "[",
            side=Side.LEFT,
        )
    return Endpoint(
        value=int(representation[:-1]),
        include=representation[-1] == "]",
        side=Side.RIGHT,
    )


def test_endpoints() -> None:
    """
    Test building endpoints.
    """
    start = build_endpoint("(0")
    assert start == Endpoint(0, False, Side.LEFT)
    assert str(start) == "(0"

    assert build_endpoint("(-10") < start
    assert build_endpoint("(0") == start
    assert build_endpoint("[0") < start
    assert build_endpoint("0]") < start
    assert build_endpoint("0)") < start

    end = build_endpoint("0]")
    assert end == Endpoint(0, True, Side.RIGHT)
    assert str(end) == "0]"

    assert build_endpoint("10)") > end
    assert build_endpoint("0)") < end
    assert build_endpoint("0]") == end
    assert build_endpoint("(0") > end
    assert build_endpoint("[0") == end

    # 0] < (0
    assert (Endpoint(0, True, Side.RIGHT) > Endpoint(0, False, Side.LEFT)) is False

    assert end != 1
    with pytest.raises(TypeError) as excinfo:
        end > 1  # pylint: disable=pointless-statement
    assert (
        str(excinfo.value)
        == "'>' not supported between instances of 'Endpoint' and 'int'"
    )


def test_is_null() -> None:
    """
    Test ``IsNull``.
    """
    assert IsNull.build([]) == IsNull()  # type: ignore
    assert IsNull().check(None) is True
    assert IsNull() != 0


def test_is_not_null() -> None:
    """
    Test ``IsNotNull``.
    """
    assert IsNotNull.build([]) == IsNotNull()  # type: ignore
    assert IsNotNull().check(None) is False
    assert IsNotNull() != 0



"""
Simple adapter for the Preset API (https://preset.io/).

This is a derivation of the generic JSON adapter that handles Preset auth.
"""

import itertools
import logging
import re
from collections.abc import Iterator
from typing import Any, Optional, cast

import jsonpath
import prison
import requests
from yarl import URL

from shillelagh.adapters.api.generic_json import CACHE_EXPIRATION, GenericJSONAPI
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Order
from shillelagh.filters import Filter
from shillelagh.lib import analyze, flatten
from shillelagh.typing import RequestedOrder, Row

_logger = logging.getLogger(__name__)

MAX_PAGE_SIZE = 100


def get_jwt_token(uri: str, access_token: str, access_secret: str) -> str:
    """
    Get JWT token from access token and access secret.
    """
    parsed = URL(uri)
    environment = parsed.host.split(".")[-3]
    api_uri = f"https://api.{environment}.preset.io/v1/auth/"

    response = requests.post(
        api_uri,
        json={"name": access_token, "secret": access_secret},
        headers={"Content-Type": "application/json"},
        timeout=60,
    )
    response.raise_for_status()
    payload = response.json()
    return cast(str, payload["payload"]["access_token"])


class PresetAPI(GenericJSONAPI):
    """
    Custom JSON adapter that handles Preset auth.
    """

    default_path = "$.payload[*]"
    cache_name = "preset_cache"

    @classmethod
    def supports(cls, uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        parsed = URL(uri)
        return (
            parsed.scheme in ("http", "https")
            and re.match(r"api\.app(-\w+)?\.preset\.io", parsed.host) is not None
        )

    def __init__(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        uri: str,
        path: Optional[str] = None,
        access_token: Optional[str] = None,
        access_secret: Optional[str] = None,
        cache_expiration: float = CACHE_EXPIRATION.total_seconds(),
    ):
        if access_token is None or access_secret is None:
            raise ValueError("access_token and access_secret must be provided")

        jwt_token = get_jwt_token(uri, access_token, access_secret)
        request_headers = {"Authorization": f"Bearer {jwt_token}"}
        super().__init__(
            uri,
            path=path,
            request_headers=request_headers,
            cache_expiration=cache_expiration,
        )


def get_urls(resource_url: str, page_size: int = MAX_PAGE_SIZE) -> Iterator[str]:
    """
    Get all URLs to download data.
    """
    baseurl = URL(resource_url)
    query = baseurl.query.get("q", "()")
    try:
        params = prison.loads(query)
    except Exception:  # pylint: disable=broad-except
        yield str(baseurl)
        return

    # assume the user knows better and keep the URL unmodified
    if "page" in params or "page_size" in params:
        yield str(baseurl)
        return

    params["page_size"] = page_size
    page = 0
    while True:
        params["page"] = page
        yield str(baseurl.with_query({"q": prison.dumps(params)}))
        page += 1


class PresetWorkspaceAPI(PresetAPI):
    """
    Adapter for Preset workspaces.
    """

    default_path = "$.result[*]"
    cache_name = "preset_cache"

    @classmethod
    def supports(cls, uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        parsed = URL(uri)
        return (
            parsed.scheme in ("http", "https")
            and parsed.host != "api.app.preset.io"
            and parsed.host.endswith(".preset.io")
        )

    def _set_columns(self) -> None:
        rows = self.get_data({}, [])

        try:
            row = next(rows)
            rows = itertools.chain([row], rows)
            column_names = list(row.keys())
        except StopIteration:
            column_names = []

        rows = itertools.islice(rows, MAX_PAGE_SIZE)
        _, order, types = analyze(iter(rows))

        self.columns = {
            column_name: types[column_name](
                filters=[],
                order=order.get(column_name, Order.NONE),
                exact=False,
            )
            for column_name in column_names
            if column_name != "rowid"
        }

    def get_data(  # pylint: disable=unused-argument, too-many-arguments, too-many-locals
        self,
        bounds: dict[str, Filter],
        order: list[tuple[str, RequestedOrder]],
        requested_columns: Optional[set[str]] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        for url in get_urls(self.uri, MAX_PAGE_SIZE):
            response = self._session.get(str(url))
            payload = response.json()
            if not response.ok:
                messages = "\n".join(
                    error.get("message", str(error))
                    for error in payload.get("errors", [])
                )
                raise ProgrammingError(f"Error: {messages}")

            rows = jsonpath.findall(self.path, payload)
            if not rows:
                break

            for i, row in enumerate(rows):
                row = {
                    k: v
                    for k, v in (row or {}).items()
                    if requested_columns is None or k in requested_columns
                }
                row["rowid"] = i
                _logger.debug(row)
                yield flatten(row)



"""
Test the holidays in-memory adapter.
"""

import datetime

from shillelagh.backends.apsw.db import connect


def test_holidays() -> None:
    """
    Test basic queries.
    """
    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = """
SELECT *
    FROM holidays
WHERE
    country='US' AND
    "date" > '2020-01-01' AND
    "date" < '2022-01-01';
"""
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("US", datetime.date(2020, 1, 20), "Martin Luther King Jr. Day"),
        ("US", datetime.date(2020, 2, 17), "Washington's Birthday"),
        ("US", datetime.date(2020, 5, 25), "Memorial Day"),
        ("US", datetime.date(2020, 7, 3), "Independence Day (observed)"),
        ("US", datetime.date(2020, 7, 4), "Independence Day"),
        ("US", datetime.date(2020, 9, 7), "Labor Day"),
        ("US", datetime.date(2020, 10, 12), "Columbus Day"),
        ("US", datetime.date(2020, 11, 11), "Veterans Day"),
        ("US", datetime.date(2020, 11, 26), "Thanksgiving"),
        ("US", datetime.date(2020, 12, 25), "Christmas Day"),
        ("US", datetime.date(2021, 1, 1), "New Year's Day"),
        ("US", datetime.date(2021, 1, 18), "Martin Luther King Jr. Day"),
        ("US", datetime.date(2021, 2, 15), "Washington's Birthday"),
        ("US", datetime.date(2021, 5, 31), "Memorial Day"),
        (
            "US",
            datetime.date(2021, 6, 18),
            "Juneteenth National Independence Day (observed)",
        ),
        ("US", datetime.date(2021, 6, 19), "Juneteenth National Independence Day"),
        ("US", datetime.date(2021, 7, 4), "Independence Day"),
        ("US", datetime.date(2021, 7, 5), "Independence Day (observed)"),
        ("US", datetime.date(2021, 9, 6), "Labor Day"),
        ("US", datetime.date(2021, 10, 11), "Columbus Day"),
        ("US", datetime.date(2021, 11, 11), "Veterans Day"),
        ("US", datetime.date(2021, 11, 25), "Thanksgiving"),
        ("US", datetime.date(2021, 12, 24), "Christmas Day (observed)"),
        ("US", datetime.date(2021, 12, 25), "Christmas Day"),
        ("US", datetime.date(2021, 12, 31), "New Year's Day (observed)"),
    ]
